/*
 * pib_control2_branch.c
 *
 * Copyright (C) 2020 Peter Belkner <pbelkner@users.sf.net>
 * Nanos gigantum humeris insidentes #TeamWhite
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2.0 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 */
#include <pib.h>

///////////////////////////////////////////////////////////////////////////////
static INT_PTR CALLBACK PibBranchProc(HWND hDlg, UINT uMsg, WPARAM wParam,
    LPARAM lParam);
static PibControl2 *pib_branch_alloc(PibControl2 *parent,
    PibControlParam2 *param, PibModel2 *model, PibControlVmt2 *vmt);
static int pib_branch_create_dialog(PibControl2 *control, HWND hDlg);
static int pib_branch_create_postproc(PibControl2 *control);
static int _pib_branch_children_create(PibControl2 *control);

///////////////////////////////////////////////////////////////////////////////
PibControl2 *pib_branch_create(HWND hDlg, PibControl2 *parent, int dialog,
    PibControlParam2 *param, PibModel2 *model, PibControlVmt2 *vmt)
{
  PibControl2 *control=pib_branch_alloc(parent,param,model,vmt);

  if (!control)
    goto e_control;

  if (dialog&&pib_branch_create_dialog(control,hDlg)<0)
    goto e_dialog;

  if (pib_branch_create_postproc(control)<0)
    goto e_child;

  ////////
  return control;
//cleanup:
e_child:
e_dialog:
#if 0 // [
e_properties:
#endif // ]
  // TODO:
  free(control);
e_control:
  return NULL;
}

void _pib_branch_destroy(PibControl2 *control)
{
  int nchildren=control->param->branch.nchildren;
  PibControl2 **cur=control->branch.children+nchildren;

  while (control->branch.children<cur) {
    --cur;

    if (*cur)
      (*cur)->vmt->destroy(*cur);
  }

  free(control);
}

static PibControl2 *pib_branch_alloc(PibControl2 *parent,
    PibControlParam2 *param, PibModel2 *model, PibControlVmt2 *vmt)
{
  int nchildren=param->branch.nchildren<1?1:param->branch.nchildren;
  size_t size=(sizeof (PibControl2))
      +(nchildren-1)*(sizeof ((PibControl2 *)NULL)->branch.children[0]);
  PibControl2 *control=calloc(1,size);

  ////////
  if (!control) {
    DMESSAGE("allocating control");
    goto e_control;
  }

  ////////
  control->vmt=vmt;
  control->parent=parent;
  control->param=param;
  control->model=model;

  ////////
  control->properties=model->vmt->properties(model);
  
  if (!control->properties) {
    DMESSAGE("missing properties");
    goto e_properties;
  }

  return control;
//cleanup:
e_properties:
  free(control);
e_control:
  return NULL;
}

static int pib_branch_create_dialog(PibControl2 *control, HWND hDlg)
{
  PibControl2 *parent=control->parent;
  PibModel2 *model=control->model;
  PibControlParam2 *param=control->param;

#if 0 // [
DVMESSAGE("\"%s\": \"%s\"",parent?parent->vmt->id:"NULL",control->vmt->id);
#endif // ]
  if (param->branch.idd) {
    // TODO:
    HWND hWndParent
        =PIB_CONTROL_TYPE_TABPAGE==control->vmt->type?GetParent(hDlg):hDlg;

    control->hDlg=CreateDialogParamW(
      param->branch.hInstance,                // HINSTANCE hInstance,
      MAKEINTRESOURCEW(param->branch.idd),    // LPCWSTR   lpTemplateName,
      hWndParent,                  				    // HWND      hWndParent,
      PibBranchProc,                          // DLGPROC   lpDialogFunc,
      (LPARAM)model                           // LPARAM    dwInitParam
    );

    if (!control->hDlg) {
      DWORD dwError=GetLastError();

      switch (dwError) {
      case ERROR_TLW_WITH_WSCHILD:
        DVMESSAGE("creating dialog (err:\"Cannot create a top-level "
            "child window\",idd:%d,type:%d)",
            param->branch.idd,control->vmt->type);
        break;
      default:
        DVMESSAGE("creating dialog (err:%lu,idd:%d)",dwError,
            param->branch.idd);
        break;
      }

      goto e_dialog;
    }

    SetWindowLongPtrW(control->hDlg,GWLP_USERDATA,(LONG_PTR)control);

    if (control->vmt->init)
      control->vmt->init(control);

#if 0 // [
    // we should run adjust when having created our children.
    if (vmt->adjust)
      vmt->adjust(control,parent);
    else {
      SetWindowPos(
        control->hDlg,      // HWND hWnd,
        NULL,               // HWND hWndInsertAfter,
        0,                  // int  X,
        0,                  // int  Y,
        0,                  // int  cx,
        0,                  // int  cy,
        SWP_NOSIZE|SWP_NOZORDER
                            // UINT uFlags
      );
    }
#else // ] [
    if (!control->vmt->adjust) {
      SetWindowPos(
        control->hDlg,      // HWND hWnd,
        NULL,               // HWND hWndInsertAfter,
        0,                  // int  X,
        0,                  // int  Y,
        0,                  // int  cx,
        0,                  // int  cy,
        SWP_NOSIZE|SWP_NOZORDER
                            // UINT uFlags
      );
    }
#endif // ]
  }
  else if (!parent) {
    DVMESSAGE("missing parent: id:\"%s\"",control->vmt->id);
    goto e_dialog;
  }
  else if (!hDlg) {
    DMESSAGE("missing hDlg");
    goto e_dialog;
  }
  else if (param->idc) {
    control->hWnd=GetDlgItem(hDlg,param->idc);

    if (!control->hWnd) {
      DVMESSAGE("control (idc:%d) not found in dialog (idd:%d)",
          param->idc,parent->param->branch.idd);
      goto e_dialog;

    }

    SetWindowLongPtrW(control->hWnd,GWLP_USERDATA,(LONG_PTR)control);

#if 0 // [
    // we should run adjust when having created our children.
    if (vmt->adjust)
      vmt->adjust(control,parent);
#endif // ]
  }
  else {
    DMESSAGE("missing idc or idd, repectively");
    goto e_dialog;
  }

  return 0;
e_dialog:
  return -1;
}

static int pib_branch_create_postproc(PibControl2 *control)
{
  PibControl2 *parent=control->parent;

  if (control->vmt->create_ex)
    control->vmt->create_ex(control,parent?parent->hDlg:NULL);

  if (_pib_branch_children_create(control)<0) {
      DMESSAGE("creating child");
      goto e_child;
  }

  // we should run adjust when having created our children.
  if (control->vmt->adjust)
    control->vmt->adjust(control,parent);

  return 0;
e_child:
  return -1;
}

void _pib_branch_rect_center(const RECT *rcParent, const RECT *rcChild,
    RECT *rcResult)
{
  long w1=rcParent->right-rcParent->left;
  long h1=rcParent->bottom-rcParent->top;
  long w2=rcChild->right-rcChild->left;
  long h2=rcChild->bottom-rcChild->top;
  long x2=rcParent->left+((w1-w2)>>1);
  long y2=rcParent->top+((h1-h2)>>1);

  rcResult->left=x2;
  rcResult->right=x2+w2;
  rcResult->top=y2;
  rcResult->bottom=y2+h2;
}

void _pib_branch_refresh(PibControl2 *control)
{
  int nchildren=control->param->branch.nchildren;
  PibControl2 **max=control->branch.children+nchildren;
  PibControl2 **cur;

  for (cur=control->branch.children;cur<max;++cur) {
    if ((*cur)->vmt->refresh)
      (*cur)->vmt->refresh(*cur);
  }
}

static int _pib_branch_children_create(PibControl2 *control)
{
  PibModel2 *model=control->model;
  PibControlParam2 *param=control->param;
  int nchildren=param->branch.nchildren;

  struct {
    PibControlParam2 *const *param;
    PibControl2 **control;
  } cur={
    .param=param->branch.children,
    .control=control->branch.children,
  };

  while (cur.param<param->branch.children+nchildren) {
    *cur.control=(*cur.param)->create(control->hDlg,control,
        (*cur.param),model);

    if (!*cur.control) {
      DMESSAGE("creating child");
      goto e_child;
    }

    ++cur.control;
    ++cur.param;
  }

  ////////
  return 0;
//cleanup:
e_child:
  while (control->branch.children<cur.control) {
    --cur.control;
    (*cur.control)->vmt->destroy(*cur.control);
    *cur.control=NULL;
  }

  return -1;
}

static INT_PTR CALLBACK PibBranchProc(HWND hDlg, UINT uMsg, WPARAM wParam,
    LPARAM lParam)
{
  PibControl2 *control=(PibControl2 *)GetWindowLongPtrW(hDlg,GWLP_USERDATA);

  switch (uMsg) {
#if 0 // [
  case WM_INITDIALOG:
DWRITELN("received \"WM_INITDIALOG\"");
    SetWindowLongPtrW(hDlg,GWLP_USERDATA,lParam);
    control=(PibControl2 *)lParam;

    if (control->vmt->init)
      control->vmt->init(control);

    break;
#endif // ]
#if 0 // [
  case WM_CLOSE:
DVWRITELN(">>> received \"WM_CLOSE\": %p",control);
#if 0 // [
    if (control&&control->vmt->close)
      control->vmt->close(control);
#endif // ]

    break;
#endif // ]
  case WM_DESTROY:
    if (control&&control->vmt) {
      DVWRITELN("WM_DESTROY:\"%s\"",control->vmt->id);
      control->vmt->close(control);
    }

    break;
  case WM_NOTIFY: {
    HWND hWndFrom=((LPNMHDR)lParam)->hwndFrom;
    PibControl2 *child=(PibControl2 *)GetWindowLongPtrW(hWndFrom,
        GWLP_USERDATA);

    if (child&&child->vmt->notify) {
DVWRITELN("WM_NOTIFY:child:%p \"%s\"",child,child->vmt->id);
      child->vmt->notify(child,wParam,lParam);
    }

    break;
  }
  case WM_COMMAND: {
    HWND hWndFrom=(HWND)lParam;
    PibControl2 *child=(PibControl2 *)GetWindowLongPtrW(hWndFrom,
        GWLP_USERDATA);

    if (child&&child->vmt->command) {
DVWRITELN("WM_COMMAND:child:%p \"%s\"",child,child->vmt->id);
      child->vmt->command(child,wParam,lParam);
    }

    break;
  }
#if 0 // [
  case WM_PIB_PROPERTIES:
    if (control&&control->vmt->postfix)
      control->vmt->postfix(control);

    break;
#endif // ]
#if 0 // [
  case WM_PIB_IPC:
    switch (wParam) {
    case PIB_CLOSE:
DWRITELN("WM_PIB_IPC/PIB_CLOSE");
      //rand_config_done(form,1,hWnd);
      break;
    case PIB_REFRESH:
DWRITELN("WM_PIB_IPC/PIB_REFRESH");
      //pib_form_post(form,hWnd,uMsg,wParam,lParam);
      break;
    default:
      break;
    }

    break;
#endif // ]
  default:
    break;
  }
//exit:
  return FALSE;
}
